Require Import List.
Require Import Arith.
Require Import LIFO.List.StrongNth.
Require Import LIFO.List.Length.
Require Import LIFO.List.Nth.
Require Import LIFO.Nat.BoundedNat.
Require Import Vector.
Require BsmlSpec.

Parameter numberOfProcessors : { n : nat | 0 < n }.

Module BsmlImplementation  <: BsmlSpec.PRIMITIVES.

  Program Definition bsp_p : nat := numberOfProcessors.

  Definition bsp_pNotZero : 0 < bsp_p := proj2_sig numberOfProcessors.
    
  Definition processor := 
    { pid: nat | pid < bsp_p }.

  Definition par (A:Type) := vector A bsp_p.

  Definition 
    get(A:Type)(v:par A)(i:processor) : 
    A :=
    proj1_sig(strongNth v i).

  Implicit Arguments get [A].

  Lemma bsp_pSMaxPid: 
    exists maxPid:nat, bsp_p = S maxPid.
  Proof.
    unfold bsp_p; destruct numberOfProcessors as [pv Hp];
      destruct pv; [ contradict Hp; omega | exists pv ;auto ].
  Qed.

  Program Definition processor0 : processor := 0.
  Next Obligation.
    destruct(bsp_pSMaxPid) as [maxPid HmaxPid];
      rewrite HmaxPid; 
        intuition.
  Qed.

  Lemma par_eq (A:Type)(v v':par A): 
    (forall (i:processor), 
      get v i = get v' i) -> v = v'.
  Proof.
    intros H; unfold get.
    rewrite  vMapNthSeq with (v:=v).
    rewrite  vMapNthSeq with (v:=v').
    apply map_ext.
    assumption.
  Qed.

  Definition processors : vector processor bsp_p := 
    seqLt bsp_p 0.

  Section Primitives.
    
    Variable A:Type.

    Program Definition 
      mkpar (f: processor -> A) : 
      { X: par A | forall i: processor, get X i = f i } :=
      map f  (seqLt bsp_p 0).
    Next Obligation.
      unfold get;
        set(H:=strongNthSeqLt0 bsp_p i);
          rewrite (proj1_sigStrongNthNth _ bsp_p i) with (d:=processor0) in H;
            rewrite (proj1_sigStrongNthNth _ bsp_p i) with (d:=f processor0);
              rewrite nthMap;
                rewrite H;
                  reflexivity.
    Defined.

    Program Definition apply (B: Type) (vf: par (A->B)) (vx: par A):
        { X: par B | forall i: processor, get X i = (get vf i) (get vx i) } := 
        map(fun (p:((A->B)*A)) => let (f,v):=p in f v)(combine bsp_p vf vx).
    Next Obligation.
      unfold get;
        rewrite strongNthMap;
          rewrite strongNthCombine with (n:=bsp_p)(v1:=vf)(v2:=vx)(i:=i);
            auto.
    Defined.

    Program Definition put (vf: par (processor -> A)):
      { X: par (processor -> A) | 
        forall i j: processor, get X i j = get vf j i } :=
      let mat := map (fun f=>map f processors) vf in 
        map
          (fun v => fun (i:processor)=>proj1_sig(strongNth v i))
          (transpose mat).
    Next Obligation.
      unfold get; rewrite strongNthMap;
        set(H:=transposeProperty); unfold Vector.get in H;
          rewrite <- H;
            repeat rewrite strongNthMap;
              unfold processors; rewrite strongNthSeqLt0 with (n:=bsp_p);
                reflexivity.
    Defined.

    Program Definition proj (v: par A) : { X: processor -> A | forall
      i: processor, X i = get v i }:=  
    (* Not really similar to operational semantics, maybe an
       alternative definition where "communications" are more visible
       would be interesting *)
      strongNth v.

  End Primitives.

  Implicit Arguments mkpar [A].
  Implicit Arguments apply [A B].
  Implicit Arguments put [A].
  Implicit Arguments proj [A].

End BsmlImplementation.