open Unix

open Bsml
open Stdlib.Base
open Stdlib.Comm

open Datatypes
open List0
open DataGenerator

let human_readable = ref false

let set_human_readable =
  human_readable := false

let compactGCsilently ()=
  begin
    print_string "SILENT GC ...";
    let verbosity = ((Gc.get()).Gc.verbose) in
      (Gc.set { (Gc.get()) with Gc.verbose = 0x000 };
       Gc.compact();
       Gc.set { (Gc.get()) with Gc.verbose =  verbosity};
      );
      print_string "END ";
  end

    
let print_separator  s =
  if (!human_readable) then
    print_string s
  else
    print_string " ; "
let gctiming = ref 0.

(** print computation time log for program [program_name] applied to data of size [size]    *)
let print_time program_name  size  time =
  begin
    let cents = int_of_float (time*.100.) in
    let sec = float_of_int (cents mod 6000)/.100. in
    let seconds = cents / 100 in
    let minutes = seconds /60 in
    let min = minutes mod 60 in
    let hours = minutes /60 in
    let h = hours mod 60 in
    let days = hours /24 in
      print_string program_name; print_separator  " computing over ";
      print_int bsp_p ; print_separator  " processor and ";
      print_int size; print_separator " data in time ";
      print_float time ; print_separator " human readable time ";
      if days > 0 then ( print_int days ; print_string "days " );
      if h>0 then ( print_int h  ; print_string "h ");
      if min>0 then ( print_int min ; print_string "min " );
      if sec>0. then (print_float sec ; print_string "sec ");
      if time < 0.01 then  (print_float time ; print_string "sec ");
      print_separator "gc time"; print_float !gctiming ;
      print_string "\n";
      flush Pervasives.stdout
  end

let print_step log =
  begin
  bsml_print  (output_string Pervasives.stderr) 0 (replicate log);
    flush Pervasives.stderr
  end

let rec string_of_list string_of_element separator = function
    | []  -> ""
    | h::[] -> string_of_element h
    | h:: t      -> (string_of_element h)^separator^
       (string_of_list string_of_element separator t)

(*
let rec string_of_Coq_list string_of_element separator = function
    | Coq_nil  -> ""
    | Coq_cons (h,nil) -> string_of_element h
    | Coq_cons (h,t)      -> (string_of_element h)^separator^
       (string_of_Coq_list string_of_element separator t)
*)

let string_of_data =  string_of_float

let print_data_coq_lst = parprint (fun l ->(print_string (string_of_list (fun c ->string_of_data c)  " ; " l)))

let print_data_lst = parprint (fun l ->(print_string (string_of_list (fun c ->string_of_data c)  " ; " l)))

let data_size_warning size =
  let local_size =size/bsp_p in
  if (local_size > 200000) then print_step ("/!\ warning /!\  local list size ("^ (string_of_int local_size) ^" exceed stack limit, risk of overflow ")


(** perform [iteration] benchmark of the function [f] applied to [data] of size [size], naming it  [program_name] in the logging output   *)
let rec benchmark  f data program_name  size iteration   =
  begin
    if iteration >0 then (
      start_timing();
      Gc.compact();
      stop_timing();
      ignore(bsml_print (fun x -> gctiming:=x) 0 (fold_direct max (get_cost())));
      start_timing();
      f size data;
      stop_timing();
      ignore(bsml_print (print_time program_name  size) 0 (fold_direct max (get_cost())));
      benchmark f data program_name  size (iteration-1);
    )
  end

(** perform [iteration] benchmark of the function [f] for each size given  in  size_list.
    data are randomly generated using balanced_random_data_par
*)
let rec benchmark_list local_data_generator f max_data program_name  size_list  iteration =
  match size_list with
    | [] -> ();
    | size::l ->(
        let data    = balanced_random_data_par local_data_generator max_data  (size/bsp_p) in
          benchmark f data program_name  size iteration ;
          benchmark_list  local_data_generator f max_data program_name  l iteration
      )
(* end *)

(* let rec coq_benchmark_list f max_data program_name  size_list  iteration = *)
(*   match size_list with *)
(*     | [] -> (); *)
(*     | size::l -> *)
(*      data_size_warning size; *)
(*      let data    = coq_balanced_random_data_par max_data  (size/bsp_p) in *)
(*        benchmark f data program_name  size iteration ; *)
(*        coq_benchmark_list f max_data program_name  l iteration *)
(* (\* end *\) *)


(* (\** small default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let small_default_benchmark f max_data program_name  :unit = *)
(*   (\* print_step ("starting small default benchmark for "^program_name^"...\n"); *\) *)
(*   let size_list = [ 1000 ; 10000; 100000; 1000000 ] in *)
(*     benchmark_list f max_data program_name  size_list 5; *)
(*     print_step ("end of default benchmark for "^program_name^".\n") *)

(* (\** small default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let coq_small_default_benchmark f max_data program_name  :unit = *)
(*   print_step ("starting small default benchmark for "^program_name^"...\n"); *)
(*   let size_list = [ 1000 ; 10000; 100000; 1000000 ] in *)
(*     coq_benchmark_list f max_data program_name  size_list 5; *)
(*     print_step ("end of default benchmark for "^program_name^".\n") *)


(* (\** default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let default_benchmark f max_data program_name   :unit = *)
(*   print_step ("starting default benchmark for "^program_name^"...\n"); *)
(*   let size_list = [ 1000 ; 10000; 100000; 1000000; 10000000; 100000000 ] in *)
(*     benchmark_list f max_data program_name  size_list 5; *)
(*     print_step ("end of default benchmark for "^program_name^".\n") *)

(* (\** default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let coq_default_benchmark f max_data program_name   :unit = *)
(*   print_step ("starting default benchmark for "^program_name^"...\n"); *)
(*   let size_list = [ 1000 ; 10000; 100000; 1000000; 10000000; 100000000 ] in *)
(*     coq_benchmark_list f max_data program_name  size_list 5; *)
(*     print_step ("end of default benchmark for "^program_name^".\n") *)


(* (\** default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let big_default_benchmark f max_data program_name   :unit = *)
(*   print_step ("starting default big benchmark for "^program_name^"...\n"); *)
(*   let size_list = [ 10000000; 50000000; 100000000 ] in *)
(*     benchmark_list f max_data program_name  size_list 5; *)
(*     print_step ("end of default benchmark for "^program_name^".\n") *)

(* (\** default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let coq_big_default_benchmark f max_data program_name   :unit = *)
(*   print_step ("starting default big benchmark for "^program_name^"...\n"); *)
(*   let size_list = [ 10000000; 50000000; 100000000 ] in *)
(*     coq_benchmark_list f max_data program_name  size_list 5; *)
(*     print_step ("end of default benchmark for "^program_name^".\n") *)


(** default benchmark perform 20 call to [f] for sizes 1000 10000 100000 and 1000000.
*)
let one_size_default_benchmark local_data_generator f max_data program_name size : unit =
  print_step ("starting  benchmark for "^program_name^"...\n");
  benchmark_list local_data_generator f max_data program_name  [size] 4;
  print_step ("end of benchmark for "^program_name^".\n")

(* (\** default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let coq_one_size_default_benchmark f max_data program_name size : unit = *)
(*   print_step ("starting  benchmark for "^program_name^"...\n"); *)
(*   let size_list = [ size ] in *)
(*     coq_benchmark_list f max_data program_name  size_list 5; *)
(*     print_step ("end of default benchmark for "^program_name^".\n") *)

(* (\** default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let test_benchmark f max_data program_name size : unit = *)
(*   let data    = balanced_random_data_par max_data  (size/bsp_p) in *)
(*     (print_step ("starting  benchmark for "^program_name^"...\n"); *)
(*       start_timing(); *)
(*       print_data_lst data; *)
(*       let v = f size data  in parprint (fun v -> print_string ("result" ^(string_of_float v))) v; *)
(*       stop_timing(); *)
(*       ignore(bsml_print (print_time program_name  size) 0 (fold_direct max (get_cost()))); *)
(*       print_step ("end of benchmark for "^program_name^".\n")) *)

(* (\** default benchmark perform 5 call to [f] for sizes 1000 10000 100000 and 1000000. *)
(* *\) *)
(* let coq_test_benchmark f max_data program_name size : unit = *)
(*   let data    = coq_balanced_random_data_par max_data  (size/bsp_p) in *)
(*     (print_step ("starting  benchmark for "^program_name^"...\n"); *)
(*       start_timing(); *)
(*       print_data_coq_lst data; *)
(*       let v = f size data in  parprint (fun v -> print_string ("result" ^(string_of_float v))) v; *)
(*       stop_timing(); *)
(*       ignore(bsml_print (print_time program_name  size) 0 (fold_direct max (get_cost()))); *)
(*       print_step ("end of benchmark for "^program_name^".\n")) *)

